@@ -2,7 +2,7 @@
 #
 # Part of the Rattle package for Data Mining
 #
-# Time-stamp: <2008-06-21 14:50:01 Graham Williams>
+# Time-stamp: <2009-08-30 10:58:34 Graham Williams>
 #
 # Copyright (c) 2009 Togaware Pty Ltd
 #
@@ -28,6 +28,9 @@
 # Date: 6 Feb 2008
 # Implements a PMML exporter for nnet objects (Neural Networks)
 #
+# 090824 gjw Instead of two output nodes should there only be one for
+# a single output node neural network?
+#
 
 # Function pmml.nnet.DataDictionary
 
@@ -142,20 +145,24 @@ pmml.nnet <- function(model,
                       app.name="Rattle/PMML",
                       description="Neural Network PMML Model",
                       copyright=NULL,
+                      transforms=NULL,
                       ...)
 {
   if (! inherits(model, "nnet")) stop("Not a legitimate nnet object")
   
   require(XML, quietly=TRUE)
   
-  ###################################################################
-  # Collect the required information. We list all variables,
-  # irrespective of whether they appear in the final model. This
-  # seems to be the standard thing to do with PMML. It also adds
-  # extra information - i.e., the model did not need these extra
-  # variables!
+  # Collect the required information.
+
+  # 090824 The number of layers is the number of hidden layers. So it
+  # is the number of layers minus 1. The nnet function will only ever
+  # build a single-hidden-layer neural network and we might expect
+  # this to be subtract 2, not subtract 1. But the representation used
+  # here feeds the normal single output layer into another layer of
+  # two nodes
   
   number.of.neural.layers <- length(model$n) - 1
+
   field <- NULL
   if (model$call[[1]] == "nnet.formula")
   {
@@ -180,58 +187,53 @@ pmml.nnet <- function(model,
       field$class[[field$name[i + 1]]] <- "numeric"
     }
   }
+
+  # 090820 Support transforms if available.
+
+  if (supportTransformExport(transforms))
+  {
+    field <- unifyTransforms(field, transforms)
+    transforms <- activateDependTransforms(transforms)
+  }
+  number.of.fields <- length(field$name)
   
-  ################################################################################
   # According to the nnet documentation:
-  # If the response in formula is a factor, an appropriate classification
-  # network is constructed; this has one output and entropy fit if the number
-  # of levels is two, and a number of outputs equal to the number of classes
-  # and a softmax output stage for more levels.
-  # If the response is not a factor, it is passed on unchanged to nnet.default.
+  #
+  # If the response in formula is a factor, an appropriate
+  # classification network is constructed; this has one output and
+  # entropy fit if the number of levels is two, and a number of
+  # outputs equal to the number of classes and a softmax output stage
+  # for more levels.  If the response is not a factor, it is passed on
+  # unchanged to nnet.default.
   #
   # However, we will actually export a network with two output neurons for binary
   # classification with a softmax output stage.
-  #
   
   normalization.method <- "none"
   skipLayers <- FALSE
   linearOutputUnits <- FALSE
   
-  if (length(model$call$skip) > 0)
-  {
-    if(model$call$skip == "T" || model$call$skip == "TRUE")
-    {
-      skipLayers <- TRUE
-    }
-  }
+  if (length(model$call$skip) && model$call$skip)
+    skipLayers <- TRUE
   if (model$nunits > model$nsunits)
-  {
     linearOutputUnits <- TRUE
-  }
   if (model$softmax)
-  {
     normalization.method <- "softmax"
-  }
   if (model$censored)
-  {
     stop("PMML does not support the censored variant of softmax!")
-  }
   
   # Levels
   
   if (field$class[[field$name[1]]] == "factor")
-  {
     field$levels[[field$name[1]]] <- model$lev
-  }
+
   factor_count <- 1
-  for (i in 1:number.of.inputs)
-  {
+  for (i in seq_len(number.of.inputs))
     if (field$class[[field$name[i + 1]]] == "factor")
     {
       field$levels[[field$name[i + 1]]] <- model$xlevels[[factor_count]]
       factor_count <- factor_count + 1
     }
-  }
   
   ##############################################################################
   # PMML
@@ -246,10 +248,8 @@ pmml.nnet <- function(model,
   
   pmml <- append.XMLNode(pmml, pmml.nnet.DataDictionary(field))
   
-  #############################################################################
   # PMML -> NeuralNetwork
-  
-  
+
   if (model$n[length(model$n)] == 1 && field$class[[field$name[1]]] == "factor")
   {
     temp <- number.of.neural.layers + 1
@@ -261,7 +261,7 @@ pmml.nnet <- function(model,
   
   if (field$class[[field$name[1]]] == "factor")
   {
-    nnet.model <- xmlNode("NeuralNetwork",
+    the.model <- xmlNode("NeuralNetwork",
                           attrs=c(modelName=model.name,
                             functionName="classification",
                             numberOfLayers=temp,
@@ -269,7 +269,7 @@ pmml.nnet <- function(model,
   }
   else
   {
-    nnet.model <- xmlNode("NeuralNetwork",
+    the.model <- xmlNode("NeuralNetwork",
                           attrs=c(modelName=model.name,
                             functionName="regression",
                             numberOfLayers=temp,
@@ -279,29 +279,39 @@ pmml.nnet <- function(model,
   # PMML -> NeuralNetwork -> MiningSchema
   
   temp = grep("as.factor", target, value = TRUE, fixed = TRUE)
-  if (length(temp) > 0)
+  if (length(temp))
   {
     tempName <- strsplit(target,"")
     endPos <- (length(tempName[[1]]) - 1)
     target <- substring(target,11,endPos)
   }
   
-  nnet.model <- append.XMLNode(nnet.model, pmml.nnet.MiningSchema(field, target))
+  the.model <- append.XMLNode(the.model, pmml.nnet.MiningSchema(field, target))
   
-  #########################################
-  #  OUTPUT
-  nnet.model <- append.XMLNode(nnet.model, pmmlOutput(field,target))
+  #  PMML -> NeuralNetwork -> Output
+
+  the.model <- append.XMLNode(the.model, pmmlOutput(field,target))
+
+  # PMML -> NeuralNetwork -> LocalTransforms
+
+  if (supportTransformExport(transforms))
+    the.model <- append.XMLNode(the.model, pmml.transforms(transforms))
   
-  ##############################################################################
   # PMML -> NeuralNetwork -> NeuralInputs
   
   neuralInputs <- xmlNode("NeuralInputs",
                           attrs=c(numberOfInputs=as.numeric(model$n[1])))
   input_count <- 1
   factor_count <- 1
-  for (i in 1:number.of.inputs)
+
+  # 090830 We need to reference everything in terms of the terms, not
+  # field, since the node numbers will be in the order of the terms,
+  # and not fields, once transforms come into play.
+  
+  for (i in seq_len(number.of.inputs))
   {
-    if (field$class[[field$name[i+1]]] == "factor")
+    # 090830 if (field$class[[field$name[i+1]]] == "factor")
+    if (terms$dataClasses[[terms$term.labels[i]]] == "factor")
     {
       number.of.values = length(model$xlevels[[factor_count]])
       usedValues <- model$xlevels[[factor_count]]
@@ -315,7 +325,7 @@ pmml.nnet <- function(model,
                                      attrs=c(id=as.numeric(input_count)))
           input_count <- input_count + 1
           
-          fieldName <- paste("derivedNI_",terms$term.labels[i],sep="")
+          fieldName <- paste("derivedNI_", terms$term.labels[i], sep="")
           fieldName <- paste(fieldName,usedValues[j],sep="")
           
           derivedFieldNode <- xmlNode("DerivedField",
@@ -341,8 +351,16 @@ pmml.nnet <- function(model,
                                  attrs=c(id=as.numeric(input_count)))
       input_count <- input_count + 1
       
-      name <- field$name[i + 1]
-      fieldName <- paste("derivedNI_",name,sep="")
+      # 090830 Use terms rather than field. The former works because
+      # its ordering will be different to foeld once we have
+      # transforms, and the the fields no longer corresponds to the
+      # terms in the model because fields will be in original variable
+      # order, whilst terms is in original variable and then transform
+      # order. But field does not have the ransform name.
+      
+#      name <- field$name[i + 1]
+      name <- terms$term.labels[i]
+      fieldName <- paste("derivedNI_", name, sep="")
       
       derivedFieldNode <- xmlNode("DerivedField",
                                   attrs=c(name=fieldName,
@@ -361,11 +379,10 @@ pmml.nnet <- function(model,
     
   }
   
-  nnet.model <- append.XMLNode(nnet.model, neuralInputs)
+  the.model <- append.XMLNode(the.model, neuralInputs)
   
   number.of.inputs <- model$n[1]
   
-  #############################################################################
   # PMML -> NeuralNetwork -> NeuralLayers
   
   wtsID <- 1
@@ -440,7 +457,7 @@ pmml.nnet <- function(model,
     
     previous.number.of.neurons <- number.of.neurons
     
-    nnet.model <- append.XMLNode(nnet.model, neuralLayerNode)
+    the.model <- append.XMLNode(the.model, neuralLayerNode)
   }
   
   # Special case for NN with 1 output neuron implementing classification
@@ -485,7 +502,7 @@ pmml.nnet <- function(model,
           
           neuralLayerNode <- append.XMLNode(neuralLayerNode, neuronNode)
           
-          nnet.model <- append.XMLNode(nnet.model, neuralLayerNode)
+          the.model <- append.XMLNode(the.model, neuralLayerNode)
           
           number.of.neurons <- number.of.neurons + 1
           
@@ -555,11 +572,11 @@ pmml.nnet <- function(model,
 		
 	}
 
-	nnet.model <- append.XMLNode(nnet.model, neuralOutputs)
+	the.model <- append.XMLNode(the.model, neuralOutputs)
 	
 	# Add to the top level structure.
 	
-	pmml <- append.XMLNode(pmml, nnet.model)
+	pmml <- append.XMLNode(pmml, the.model)
 	
 	return(pmml)
 }
